/*************************************************************************
**	C version of CIO
**	Copyright (c) 1991 by Jim Patchell
**
**	This is the Central Input Output device handler.  This is probably
**	a lame attempt to make access to various I/O devices a little more
**	orthagonal.  Maybe, maybe not.
**
**  The interface function, cio only requires one parameter, but accepts
**	a variable number of parameters depending on the type of function
**	it is that is being executed.
**
**	The functions that can be executed are:
**
**		add handler
**		open
**		close
**		get
**		get record (read)
**		put
**		put record (write)
**		status
**		special command (xio)
**
**
**************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "cio.h"
#include "task.h"
#include "queue.h"

/*
	cio global variables
*/

static IOCB Iocb[32];	/*	space for 32 I/O channels	*/
static HTABS htabs[20];	/*	space for 20 I/O devices	*/
static int last_dev = 0;	/*	last device entry in table	*/

static free_iocb(int handle);
static find_free_iocb(void);
static dev_search(int handle);

static Wait *IocbEvent = NULL;

extern "C" long cio(va_list argp)
{
	/***********************************************************************
	** this is the Central I/O function (cio)
	** This is where all of the high level IO access is handled.
	**
	** Parameter:
	** argp.....pointer to a list of arguments
	** ReturnValue:
    ** Returns a long value depending on the function
	***********************************************************************/
	int handle;
	int a;		/*	temporary data storage	*/
	long b;		//another temporary data storage
	IOCB *iocb;
	int cmd;

	cmd = va_arg(argp,int);		//get function command
	iocb = Iocb;
	switch(cmd)
	{
		case ADD_HANDLER:
			/**************************************************************
			** This command is used to insert handler vector into HTABS
			** called as: cio(cmd,devname,table)
			** Also Called as:AddHandler(const char *dev_name,const H_JVEC *jump_table);
			**
			** Parameter:
			** cmd......Will have the value ADD_HANDLER
			** devname..Points to a string that contains the device name
			** table....Points to an H_JVEC structure that contains the
			**          driver entry points
			** ReturnValue:
			** Always returns 0.
			**************************************************************/
			IocbEvent->Pend();
			htabs[last_dev].dev_name = va_arg(argp,char *);	/*	device name	*/
			htabs[last_dev].HtabsEntry = va_arg(argp,H_JVEC *);	/*	address of table	*/
			++last_dev;
			IocbEvent->Post();		//unblock add handler
			break;
		case OPEN:
			/*************************************************************
			** This command is used to open up the device.
			** called as: cio(int cmd,char *dev_name,int mode);
			** also called as:Open(char *dev_name,int mode);
			**
			** Parameter:
			** cmd.......Will have the value OPEN.
			** dev_name..Name of device to open
			** mode......Mode to open device as (Read/Write).
			** ReturnValue:
			** Returns handle to device (0 or positive), or negative on error
			*************************************************************/
			IocbEvent->Pend();
			if((handle = find_free_iocb()) == -1)
			{
				IocbEvent->Post();
				return(CIO_NO_HANDLES);
			}
			/*
				get open parrameters off of stack
				passed as cio(cmd,name,mode)
			*/
			iocb += handle;
			iocb->dev_name = va_arg(argp,char *);
			iocb->mode = va_arg(argp,int);
			if(( iocb->ichid = dev_search(handle)) < 0)
			{
				IocbEvent->Post();
				return(CIO_NO_DEVICE);	/*	could not find device handle	*/
			}
			/* we have laid claim to the IOCB, release iocbevent	*/
			IocbEvent->Post();
			if((a = (*htabs[iocb->ichid].HtabsEntry->openv)(iocb,argp)) < 0)
			{
				handle = a;		//return ERROR
				iocb->ichid = -1;	//free up IOCB again
			}
			return(handle);
		case CLOSE:
			/****************************************************************
			** This command is used to terminate (release) access to an
			** IO device.
			** called as: cio(int cmd,int handle);
			** Also Called as:Close(int handle);
			**
			** Parameter:
			** cmd...........has the value of CLOSE
			** handle........handle of device, aquired by using OPEN
			** ReturnValue:
			** Returns 0 on success, negative on error
			****************************************************************/
			/*
				get close parrameters off of stack
				passes as: cio(cmd,handle);
			*/
			handle = va_arg(argp,int);
			iocb += handle;
			if(handle > MAX_HANDLE || handle < 0)
				return(CIO_INVLD_HNDL);
			if(iocb->ichid < 0)
				return(CIO_NOT_OPEN);
			(*htabs[iocb->ichid].HtabsEntry->closv)(iocb,argp);
			return(free_iocb(handle));
		case GETC:
			/***************************************************************
			** This command is used to get a character from a device
			** called as: cio(int cmd, int handle);
			** Also called as:Getc(int handle);
			**
			** Parameter:
			** cmd.....Has the value of GETC
			** handle..Handle of device to get data from
			** Return Value:
			** Returns character (positive), or error (negative) value.
			** It should be noted that the return value is a long.
			** Most devices generally deal in bytes.  The byte value is
			** not sign extended, so a valid return value is always positive
			***************************************************************/
			/*
			** Get character from handler
			** called as:  cio(GETC,handle)
			*/
			handle = va_arg(argp,int);
			if(handle > MAX_HANDLE || handle < 0)
				return(CIO_INVLD_HNDL);
			iocb += handle;
			if(iocb->ichid < 0)
				return(CIO_NOT_OPEN);
			if((cmd & iocb->mode) == 0)
				return(CIO_WRONLY);
			//
			// look up function
			//
			return ( (*htabs[iocb->ichid].HtabsEntry->getv)(iocb,argp)  );
		case READ:
			/**************************************************************
			** Read a block of data from a device
			** called as: cio(READ,handle,buffer,count)
			** Also called as:Read(int handle,char *buff,long count)
			**
			** parameter:
			** cmd......This value will be READ
			** handle...Handle of device to be read.
			** buffer...character array where data will be stored
			** count....number of characters to read
			** ReturnValue:
			** Returns the number of characters read or negative on error
			***************************************************************/
			handle = va_arg(argp,int);
			if(handle > MAX_HANDLE || handle < 0)
				return(CIO_INVLD_HNDL);
			iocb += handle;
			if(iocb->ichid < 0)
				return(CIO_NOT_OPEN);
			if((cmd & iocb->mode) == 0)
				return(CIO_WRONLY);
			//
			//look up function
			//
			return (*htabs[iocb->ichid].HtabsEntry->readv)(iocb,argp);
		case PUTC:
			/**************************************************************
			** This command outputs a character to the device
			** Called As: cio(PUTC,handle,data)
			** Also called as:Putc(int handle,int c);
			**
			** Parameter:
			** cmd........This value will be PUTC
			** handle.....Handle of device to send character to
			** data.......Character to send.
			** ReturnValue:
			** Returns 0 on success, negative value on failure
			**************************************************************/
			handle = va_arg(argp,int);
			if(handle > MAX_HANDLE || handle < 0)
				return(CIO_INVLD_HNDL);
			iocb += handle;
			if(iocb->ichid < 0)
				return(CIO_NOT_OPEN);
			if((iocb->mode & cmd) == 0)
				return(CIO_RDONLY);	/*	read only error	*/
			a = va_arg(argp,int);
			//
			//look up function
			//
			return( (*htabs[iocb->ichid].HtabsEntry->putv)(iocb,a,argp));
		case WRITE:
			/**************************************************************
			** This command writes a block of data to the device
			** Called As: cio(WRITE,handle,buffer,count);
			** Also Called As:Write(int handle,char *buff,long count);
			**
			** Parameter:
			** cmd.......this value is set as WRITE.
			** handle....Handle of device to write block to
			** buff......Pointer to array of characters to write
			** count.....Number of characters to write to device
			** ReturnValue:
			** returns either an error or actual number of bytes written
			**************************************************************/
			handle = va_arg(argp,int);
			if(handle > MAX_HANDLE || handle < 0)
				return(CIO_INVLD_HNDL);
			iocb += handle;
			if(iocb->ichid < 0)
				return(CIO_NOT_OPEN);
			if((iocb->mode & cmd) == 0)
				return(CIO_RDONLY);	/*	read only error	*/
			return( (*htabs[iocb->ichid].HtabsEntry->writev)(iocb,argp));
		case STATUS:
			/************************************************************
			** This command returns the device status
			** called as: cio(cmd,handle,buffer,count,aux)
			** called as: cio(cmd,-1,dev_name,buffer,count,aux)
			** Also called as:Status(int handle,char *buff,long count,int aux);
			** Or called as  :StatusO(char *dev_name,char *buff,long count,int aux);
			**
			** Parameter:
			** cmd.......this value is set as STATUS.
			** handle....handle of device.  If -1, use next param as device name
			** dev_name..pointer to string naming device to open (if handle < 0)
			** buff......pointer to buffer to put status information
			** count.....size of buffer
			** aux.......aux information to determine what to return status of
			**
			*************************************************************/
			handle = va_arg(argp,int);
			if(handle < 0)
			{	/*	this is immeadiate operation	*/
				IocbEvent->Pend();
				if((handle = find_free_iocb()) == -1)
				{
					IocbEvent->Post();
					return(CIO_NO_HANDLES);	/* open up an IOCB	*/
				}
				iocb += handle;
				iocb->dev_name = va_arg(argp,char *);
				if(( iocb->ichid = dev_search(handle)) < 0)
				{
					IocbEvent->Post();
					return(CIO_NO_DEVICE);	/*	could not find device handle	*/
				}
				/* we have grabed this IOCB	*/
				IocbEvent->Post();
				b = (*htabs[iocb->ichid].HtabsEntry->statv)(iocb,argp);
				free_iocb(handle);	/*	close iocb	*/
				return(b);
			}
			else
			{	/*	valid handle for already open iocb	*/
				iocb += handle;
				if(handle > MAX_HANDLE || handle < 0)
					return(CIO_INVLD_HNDL);
				return((*htabs[iocb->ichid].HtabsEntry->statv)(iocb,argp) );
			}
		default:
			/**************************************************************
			** All other commands are processed here (special commands)
			**
			** called as:  cio(cmd,handle,dev_name,buffer,count,aux)
			** Also Called as:Xio(int cmd,int handle,char *dev_name,char *buff,long count,int aux,...);
			**
			** parameter:
			** cmd.......special command to handle
			** handle....handle of device.  If handle < 0, use dev_name to open device
			** dev_name..name of device to open.  If handle >= 0, ignore.
			** buff......data buffer pointer (if needed)
			** count.....size of data buffer (if needed)
			** aux.......aux data for handler
			** ReturnValue:
			** Returns negative value on error.  0 or positive, depending of function, on succes
			**
			**************************************************************/
			if((handle = va_arg(argp,int)) < 0)
			{
				/*
					this is an immediate operation
				*/
				IocbEvent->Pend();
				if((handle = find_free_iocb()) == -1)
				{
					IocbEvent->Post();
					return(-1);	/* open up an IOCB	*/
				}
				iocb += handle;
				iocb->dev_name = va_arg(argp,char *);
				if(( iocb->ichid = dev_search(handle)) < 0)	/*	locate handler entry table	*/
				{
					IocbEvent->Post();
					return(-1);	/*	could not find device handle	*/
				}
				/*	we have grabbed the IOCB	*/
				IocbEvent->Post();
				b = (*htabs[iocb->ichid].HtabsEntry->specv)(cmd,iocb,argp);
				free_iocb(handle);
				return(b);
			}
			else
			{
				/*
				** iocb is already open
				*/
				iocb += handle;
				va_arg(argp,char *);	//get rid of dummy device name
				return((*htabs[iocb->ichid].HtabsEntry->specv)(cmd,iocb,argp));
			}
	}	/*	end switch cmd	*/
	return 0l;	//everything is peachy keen
}

static dev_search(int handle)
{
	/*********************************************************************
	** This function is used to find a device name in the driver lookup
	** table.
	**
	** parameter:
	** handle.....handle of device.  dev_name is stored in the IOCB prior
	**            to this function being called.
	** ReturnValue:
	** Returns Index into device table, negative on error
	*********************************************************************/
	int i;	//this value will be returned, index into device table
	HTABS *ht;	//pointer to device table
	IOCB *iocb;	//pointer to IOCB for device
	char *n,un[4];		/*unit numbers up to 999	*/
	int l;		/*	legnth of device name	*/

	iocb = Iocb + handle;	//calculate address of IOCB for this device
	//-----------------------
	// Isolate the device name
	// from any unit numbers
	// that might be present
	//-----------------------
	l = strcspn(iocb->dev_name,":123456789");	/*	we need to find what is not unit number	*/
	//-----------------------
	// locate last entry in
	// the device table
	//-----------------------
	ht = htabs + last_dev - 1;									/*	point to handler table	*/
	//-----------------------
	//search the device table
	//search from the end so
	//that if no device is
	//found, the index "i"
	//will end up with a neg
	//value.
	//-----------------------
	for(i=last_dev -1;(i > -1) && strncmp(iocb->dev_name,ht->dev_name,l) ;--i,--ht);
	//-----------------------
	//if the index "i" is
	//positive, complete
	// processing
	//-----------------------
	if(i >= 0)
	{
		/*
		** ok, now we need to locate the unit number
		*/
		n = iocb->dev_name + l;	//points to end of device name
		if((l = strcspn(n,":")) == 0)	//if the end char is ':', then
		{
			iocb->devnum = 1;		/*	default unit number	*/
		}
		else                        //else end is unit number
		{
			strncpy(un,n,l);        //copy unit number
			un[l] = '\0';
			iocb->devnum = atoi(un);	/*	convert unit number to int	*/
		}
	}
	return(i);	//return index to CIO
}

static find_free_iocb(void)
{
	/**********************************************************************
	** This function is used to find an unused IOCB block
	**
	** Parameter:
	** <none>
	** ReturnValue:
	** returns postive number if free IOCB found (handle), or negative
	** on error.
	**********************************************************************/
	int i;
	IOCB *iocb;

	for(i=0,iocb = Iocb;i<32;++i,++iocb)
	{
		if(iocb->ichid == -1)	//is IOCB in use?
			return(i);
	}
	return(-1);	//no IOCB's found.
}

static free_iocb(int handle)
{
	/***********************************************************************
	** This function is used to release an IOCB (close)
	**
	** parameter:
	** handle....device handle of block to free
	** ReturnValue:
	** this function always returns 0
	***********************************************************************/
	Iocb[handle].ichid = -1;
	return(0);
}

extern "C" void init_iocb(void)
{
	/**********************************************************************
	** This function initialized everything in this module
	**
	** parameter:
	** <none>
	** ReturnValue:
	** <none>
	**********************************************************************/
	int i;
	IOCB *iocb;

	IocbEvent = new Wait(1,"CIO_BLOCKING_SEMAPHORE");	//create blocking semaphore for CIO
	for(i=0,iocb = Iocb;i<32;++i,++iocb)
	{
		iocb->ichid = -1;
		iocb->p = (void *)0;
	}
	last_dev = 0;
}
