/************************************************************************
**
** This is the main module for the Midi Controller board.
** This is where we initialize the multitasking system, hardware, and
** start things rolling
**
** Copyright (c) 2002 by Jim Patchell
**
************************************************************************/

#define RS232_BUILD

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cio.h"
#include "task.h"
#include "queue.h"
#include "timer.h"
#include "rs232.h"

extern "C" void DisableInterrupt(void);
static void StartMultitasking(void);
int ConsolHandle;
int MidiHandle;
TSemaphore *PortBlock;
TSemaphore *NoteOn[32];
TSemaphore *MidiRT;



#define PORT_A 0xd0000
#define N_NOTES	30

char PortAShadows[4];

void main()
{
	InitOS();
	TBlock();			//prevent task swaping until ready
	init_iocb();		/*	initialize CIO	*/
	R_init();			/*	initialize first RS232 port	*/
	init_timer();		/*	initialize real time clock	*/
	StartMultitasking();
	while(1);			/*	just stop everything	*/
}

static void SetPortBit(int bit,int state)
{
	/*******************************************************************
	** This function is used to set a bit in the solanoid port
	** This port is 32 bits wide
	** This function is protected so that multiple tasks may have
	** access
	** bit selects the bit
	** state determins on or off
	*******************************************************************/
	int port = (bit & 0x1f) >> 3;	//port index
	PortBlock->Pend();	//grab port
	if(state)
		PortAShadows[port] |= 1 << (bit & 0x07);
	else
		PortAShadows[port] &= ~(1 << (bit & 0x07));
	*( (char *)(PORT_A + port) ) = PortAShadows[port];	//write to port
	PortBlock->Post();	//release port
}


static void NoteOnTask(int note)
{
	/*
	** this task is used more than one.  The value of note determines
	** which key is going to be activated
	*/
	char *s = new char[64];
	int l;

//	l = sprintf(s,"Task %d Active\r\n",note);
//	Write(ConsolHandle,s,l);
	while(1)
	{
		NoteOn[note]->Pend();	//wait for note on message
		SetPortBit(note,1);	//turn not on
		l = sprintf(s,"Note %d On\r\n",note);
		Write(ConsolHandle,s,l);
		TDelay(5);			//wait a little bit
		SetPortBit(note,0);	//turn off note
	}
}

#define MIDICLOCK	0
#define MIDISTART	1
#define MIDISTOP	2
#define MIDIRESET	3

static void MidiRealTimeTask(void)
{
	char *s = new char[64];
	int c;
	int MidiClock= 23;

	while(1)
	{
		c = MidiRT->Pend();	//wait for message
		switch(c)
		{
			case MIDICLOCK:
				--MidiClock;
				if(MidiClock == 0)
				{
					MidiClock = 23;
					Putc(ConsolHandle,'C');
				}
				break;
			case MIDISTART:
				Putc(ConsolHandle,'S');
				break;
			case MIDISTOP:
				Putc(ConsolHandle,'P');
				break;
			case MIDIRESET:
				Putc(ConsolHandle,'R');
				break;
		}
	}
}

enum {IDLE,NOTE0,NOTE1};

#define SYSEX			0xf0
#define SYSCOMUNDEF		0xf1
#define SYSCOMSONGPOS	0xf2
#define SYSCOMSONGSEL	0xf3
#define SYSCOMUNDEF1	0xf4
#define SYSCOMUNDEF2	0xf5
#define SYSCOMTUNEREQ	0xf6
#define SYSEXEND		0xf7
#define SYSRTCLOCK		0xf8
#define SYSRTUNDEF		0xf9
#define SYSRTSTART		0xfa
#define SYSRTCONTINUE	0xfb
#define SYSRTSTOP		0xfc
#define SYSRTUNDEF1		0xfd
#define SYSRTACTIVESEN	0xfe
#define SYSRTRESET		0xff

#define NOTEON	0x90
#define NOTEOFF	0x80

static void MidiTask(void)
{
	//----------------------------------------------------------------
	// This task is used to process incoming midi information
	//----------------------------------------------------------------
	int LastCommand=-1;
	int state;
	int c,l,cmd,chan,note;
	char *s = new char[128];

	Xio(RS232_SET_GETEOT,ConsolHandle,(char *)0,(char *)0,0l,-1);	//no eot character
	Xio(RS232_SET_PUTEOT,ConsolHandle,(char *)0,(char *)0,0l,-1);	//no eot character
	Xio(RS232_SET_GETEOT,MidiHandle,(char *)0,(char *)0,0l,-1);	//no eot character
	Xio(RS232_SET_PUTEOT,MidiHandle,(char *)0,(char *)0,0l,-1);	//no eot character
	l = sprintf(s,"Midi Task is ACTIVE\r\n");
	Write(ConsolHandle,s,l);
	while(1)	//this is a task, loop forever
	{
		c = Getc(MidiHandle);
		if(c & 0x080)	//is it a command?
		{
			if(c < 0x0f0)
			{
				//process this stuff, dispose of everything else
				cmd = c & 0xf0;
				if(cmd == NOTEON)
					state = NOTE0;
			}
			else
			{
				switch(c)
				{
					case SYSEX:
						break;
					case SYSCOMUNDEF:
						break;
					case SYSCOMSONGPOS:
						break;
					case SYSCOMSONGSEL:
						break;
					case SYSCOMUNDEF1:
						break;
					case SYSCOMUNDEF2:
						break;
					case SYSCOMTUNEREQ:
						break;
					case SYSEXEND:
						break;
					case SYSRTCLOCK:
						MidiRT->Post(MIDICLOCK);
						break;
					case SYSRTUNDEF:
						break;
					case SYSRTSTART:
						MidiRT->Post(MIDISTART);
						break;
					case SYSRTCONTINUE:
						break;
					case SYSRTSTOP:
						MidiRT->Post(MIDISTOP);
						break;
					case SYSRTUNDEF1:
						break;
					case SYSRTACTIVESEN:
						break;
					case SYSRTRESET:
						MidiRT->Post(MIDIRESET);
						break;

				}
			}

		}
		else
		{
			if(cmd == NOTEON)
			{
				switch(state)
				{
					case NOTE0:
						note = c - 43;
						if(note < 0) note = 0;
						else if(note > 29) note = 29;
						state = NOTE1;
						break;
					case NOTE1:
						if(c > 0)
							NoteOn[note]->Post();
						state = NOTE0;
						break;
				}
			}
		}

	}	//end of while(1) loop
}

static void IdleTask(void)
{
	volatile int i;

	while(1)
		++i;
}

static void StartMultitasking(void)
{
	/*
	** this is it, the REAL test
	**
	** Just do the same thing that was done in the original test
	*/
	TCB *t;
	int i;
	/*
	** create TASKS
	*/

	t = CreateTask(IdleTask,2048,1,"IdleTask");
	ActiveTasks->Insert(t);

	char *s = new char[32];
	for(i=0;i<30;++i)	//create note on tasks
	{
		sprintf(s,"NoteTask%d",i);
		t = CreateTaskP(NoteOnTask,256,10,s,i);
		ActiveTasks->Insert(t);
	}
	MidiRT = new TSemaphore(0,TSEMAPHORE_MODE_WAIT,"MidiRT");
	t = CreateTask(MidiRealTimeTask,256,15,"RealTimeMessage");
	ActiveTasks->Insert(t);
	t = CreateTask(MidiTask,512,12,"MidiTask");
	ActiveTasks->Insert(t);
	PortBlock = new TSemaphore(1,TSEMAPHORE_MODE_WAIT,"PORT_A_BLOCKER");

	for(i=0;i<30;++i)	//create not on tasks
	{
		sprintf(s,"NoteSem%d",i);
		NoteOn[i] = new TSemaphore(0,TSEMAPHORE_MODE_WAIT,s);
	}
	delete[]s;
	//
	// Open up channels to standard devices
	//

	ConsolHandle = Open("R1:",(READ_ONLY | WRITE_ONLY));
	MidiHandle = Open("R2:",(READ_ONLY | WRITE_ONLY));
	Xio(MEM_PROTECT,ConsolHandle,(char *)0,(char *)0,0l,0);
	Xio(SET_ACTIVE,ConsolHandle,(char *)0,(char *)0,0l,0);
	Xio(SET_ERROR,ConsolHandle,(char *)0,(char *)0,0l,0);
	DisableInterrupt();
	TRelease();		//release scheduler
	Start();		//start multitasker
}
