#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "task.h"
#include "queue.h"
#include "cio.h"

//**************************************************************************
//
// This file contains the definitions for the various message queues that
// are needed by the teletrac spindle controller board.
// All of these queues are derived from TSemaphore in task.h
//
//
//**************************************************************************


//--------------------------------------------------------------------------
// Class Definitions
// Class Wait
// Not much different from TSemaphore.  It just doesn't support timeouts
//--------------------------------------------------------------------------

Wait::Wait(int count,char *name):TSemaphore(count,TSEMAPHORE_MODE_WAIT,name)
{
	value = 0;
}

Wait::~Wait()
{
}

int Wait::Pend(int timeout)
{
	int retval = TSemaphore::Pend(timeout);
	if(retval > 0) retval = value;
	return retval;
}

int Wait::Post(int v)
{
	TCB *t;
	int sr;

	value = v;
	return TSemaphore::Post(v);
}

void *Wait::operator new(size_t size)
{
	void *t = malloc(size + 128);
	((Wait *)t)->name = ((char *)t) + size;
	return t;
}

void Wait::operator delete(void *t)
{
	free(t);
}

//-------------------------------------------------------------------------
// Event Queue
//
// This is different from the Message Queue in that this is used when a
// Timeout in not desired.  This will help cut down on the overhead
// involved in using a semaphore.  Only one task can wait on an Event Queue,
// but many tasks can post to one.
//-------------------------------------------------------------------------

EventQueue::EventQueue(int size,char *name):TSemaphore(0,TSEMAPHORE_MODE_WAIT,name)	//constructor function
{
	Buff = new char[size];
	Head = 0;
	Tail = 0;
	nChars = 0;
	Size = size;
}

EventQueue::~EventQueue()
{
	if(Buff)
		delete[] Buff;
}

void *EventQueue::operator new(size_t size)
{
	void *t = malloc(size + 128);
	((EventQueue *)t)->name = ((char *)t) + size;
	return t;
}

void EventQueue::operator delete(void *t)
{
	free(t);
}

int EventQueue::Post(int v)
{
	//this function will post the semaphore, but does not alter the queue
	//the value v will be returned by pend
	return TSemaphore::Post(v);
}


int EventQueue::PostMessage(int *d,int c)
{
	//this member function both writes data into the queue and performs a post
	TCB *t;
	int sr;

	sr = EnterCritical();
	if(EventCount < 32766)	//is it going to overflow??
	{
		//-----------------------
		//write data into queue
		//-----------------------
		while(c)
		{
			if(nChars < Size)
			{
				Buff[Head++] = (char)*d++;
				if(Head >= Size)
					Head = 0;		//wrap head pointer
				++nChars;			//increment number of characters
			}
			else
			{
				ExitCritical(sr);
				return EVENT_BUFFFULL;
			}
			--c;
		}
		if(EventCount++ >= 0)	//increment semaphore
		{
			ExitCritical(sr);
		}
		else
		{
			if(task_h)				//is there really a task waiting?
			{
				t = task_h;		//get first task in
				task_h = t->next;       //remove from list
				t->next = (TCB *)0;
				t->status = EVENT_NOERROR;			//return this value
				t->pending = 0;
				ActiveTasks->Insert(t);	//put task back onto active list
				ExitCritical(sr);
				Schedule();
			}
			else
			{
				ExitCritical(sr);
			}
		}
		return EVENT_NOERROR;
	}
	else
	{
		ExitCritical(sr);
		return(EVENT_OVERFLOW);
	}
}

int EventQueue::Pend(int v)
{
	return TSemaphore::Pend(v);
}

int EventQueue::Get(void)
{
	int retval;
	int sr;

	if(nChars)
	{
		sr = EnterCritical();
		retval = Buff[Tail++];
		if(Tail >= Size)
			Tail = 0;		//wrap tail pointer
		--nChars;
		ExitCritical(sr);
	}
	else
		retval = -1;
	return retval;
}

int EventQueue::Put(int c)
{
	int retval = c;
	int sr;

	sr = EnterCritical();
	if(nChars < Size)
	{
		Buff[Head++] = (char)c;
		if(Head >= Size)
			Head = 0;		//wrap head pointer
		++nChars;			//increment number of characters
	}
	else
		retval = -1;		//could not put character into buffer
	ExitCritical(sr);
	return retval;
}

int EventQueue::QueueStatus(void)	//number of char in queue
{
	return nChars;
}

int EventQueue::QueueSpace(void)	//how much space left in queue
{
	return Size - nChars;
}

void EventQueue::Kill(void)
{
	int sr;

	sr = EnterCritical();
	nChars = 0;
	Tail = Head;
	ExitCritical(sr);
}

