;***********************************************************************
;
; Startup code for Spindle Controller
; Initialize memory and call main.
; Contains vectors for interrupts
;
;***********************************************************************
	XDEF	START,__brkp,__brksz,__exit
	XREF	STKTOP,DATA,_main
	XREF	MALLOC_SIZE,MALLOC_START

;***********************************************************************
;	If C++ style startup code is not desired, then 
;	define CPLUSPLUS to 0
;***********************************************************************
CPLUSPLUS	= 1

;***********************************************************************
;	Define vector area
;***********************************************************************
	SECTION vectors
__vectors	DS.L	256	;256 vectors at start of ram

;***********************************************************************
;	Define variables to track memory allocations for mbrk().
;***********************************************************************
	SECTION	data
__brkp	DC.L	MALLOC_START			; point to available memory
__brksz	DC.L	MALLOC_SIZE			; number of available bytes

;***************************************************************
;	The reset vector will point to this code (START).
;	By the time we get here, A7 will have been initialized.
;***************************************************************
	SECTION	code
	.STK	"none"			; terminate call chain for -OG
START	MOVE.L	#STKTOP,A7		; set the stack pointer
	MOVE.L	#0,A6			; end of stack frame chain

	;***********************************************************
	;		Zero out uninitialized RAM.
	;***********************************************************
	MOVE.L	#`BASE(ram),A1		; A1 = base of region ram
	MOVE.L	#`SIZE(ram),D0		; D0 = size of region ram
	LSR.L	#2,D0			; compute size in longs
	BRA	ZDBF			; enter a fast loop
  ZLP:	CLR.L	(A1)+			; clear four bytes at a time
  ZDBF:	DBF	D0,ZLP			; up to 256K in inner loop
	SUB.L	#$10000,D0		; rest in outer loop
	BHS	ZLP

	;***********************************************************
	;		Initialize other RAM from ROM.
	;***********************************************************
	MOVE.L	#DATA,A0		; A0 = ROM base of region data
	MOVE.L	#`BASE(data),A1		; A1 = RAM base of region data
	MOVE.L	#`SIZE(data),D0		; D0 = size of region data
	LSR.L	#2,D0			; compute size in longs
	BRA	IDBF			; enter a fast loop
  ILP:	MOVE.L	(A0)+,(A1)+		; move four bytes at a time
  IDBF:	DBF	D0,ILP			; up to 256K in inner loop
	SUB.L	#$10000,D0		; rest in outer loop
	BHS	ILP
	;
	;***********************************************************
	;		Zero out Heap
	;***********************************************************
	move.l	__brkp,a1		;a1 = base of heap
	move.l	__brksz,d0		;d0 = size of heap
	lsr.l	#2,d0			;compute size in longs
	move.l	#$deadbeaf,d1	;load in data
	bra	zhpdbf
zhp:
	move.l	d1,(a1)+			;clear four bytes at a time
zhpdbf:
	dbf	d0,zhp				;up to 256k in inner loop
	sub.l	#$10000,d0		;rest in outer loop
	bhs	zhp

	;
	;***********************************************************
	;		Initialize RAM vectors from ROM
	;***********************************************************
	;
;	XREF	EXCEPTION_VECTORS
	MOVE.L	#exception_vecs,A0	;point A0 to ROM image of exception vector data
	MOVE.L	#`BASE(vectors),A1	;point A1 to RAM base of region vectors
	MOVE.L	#`SIZE(vectors),D0	;set D0 to size of region vectors
	LSR.L	#2,D0			;compute the size in longs
	BRA	EDBF			;enter a fast loop
ELP:	MOVE.L	(A0)+,(A1)+		;move four bytes at a time
EDBF:	DBF	D0,ELP
	;
	;***********************************************************
	;		Initialize memory allocator.
	;***********************************************************
;	 MOVE.L	#`BASE(malloc),D0	; address of malloc region
;	MOVE.L	#`SIZE(malloc),D1	; size of malloc region
;    .IF "ptrd"?"2"
;	SWAP	D0			; accommodate 2-byte C "pointers"
;    .ENDIF
;    .IF "long"?"2"
;	SWAP	D1			; accommodate 2-byte C "longs"
;    .ENDIF
;	MOVE.L	D0,__brkp		; variables referenced by mbrk()
;	MOVE.L	D1,__brksz
;
 .IF CPLUSPLUS
	;***********************************************************
	;		Call any C++ initializer thunks.
	;***********************************************************
	MOVE.L	#`BASE(init),A5		; A5 = base of region init
	MOVE.L	#`SIZE(init),D7		; D7 = size of region init
    .IF "ptrf"?"2"
	LSR.L	#1,D7			; number of initializer thunks
    .ELSE
	LSR.L	#2,D7			; number of initializer thunks
    .ENDIF
	BRA	NDBF			; enter loop
  NLP:
    .IF "ptrf"?"2"
	MOVE.W	(A5)+,A0		; get address of thunk
    .ELSE
	MOVE.L	(A5)+,A0		; get address of thunk
    .ENDIF
	CMP.W	#0x4E71,A0		; NOP may pad region
	BEQ	NDBF
	JSR	(A0)			; call initializer thunk
  NDBF:	DBF	D7,NLP			; up to 256K in inner loop
	SUB.L	#$10000,D7		; rest in outer loop
	BHS	NLP
 .ENDIF
	;***********************************************************
	;		Invoke main() with no arguments.
	;***********************************************************
	JSR	_main			; any return value is "int" in D0
	;***********************************************************
	;		Call any C++ exit thunks.
	;***********************************************************
__exit
 .IF CPLUSPLUS
	MOVE.L	#`BASE(exit),A5		; A5 = base of region exit
	MOVE.L	#`SIZE(exit),D7		; D7 = size of region exit
	ADD.L	D7,A5			; reverse order
    .IF "ptrf"?"2"
	LSR.L	#1,D7			; number of exit thunks
    .ELSE
	LSR.L	#2,D7			; number of exit thunks
    .ENDIF
	BRA	XDBF			; enter loop
  XLP:
    .IF "ptrf"?"2"
	MOVE.W	-(A5),A0		; get address of thunk
    .ELSE
	MOVE.L	-(A5),A0		; get address of thunk
    .ENDIF
	CMP.W	#0x4E71,A0		; NOP may pad region
	BEQ	XDBF
	JSR	(A0)			; call exit thunk
  XDBF:	DBF	D7,XLP			; up to 256K in inner loop
	SUB.L	#$10000,D7		; rest in outer loop
	BHS	XLP
 .ENDIF

  DONE:	BRA	DONE			; infinite loop if main ever returns

;***********************************************************************
;	OTHER EXCEPTION VECTORS: to supervisor data space at address 8,
;	or 8 bytes beyond where the vector base register will point.
;	This table is commented out because no actual interrupt routines
;	are provided.
;***********************************************************************
	SECTION	vects
	XREF buss_error,address_error
	XREF illegal,zero_div,chk_instr,trapv_instr,privilege
	XREF trace,line_a,line_f,reserved,unitialized
	XREF spurious,autovector,trap_instr,user_vecs
	XREF _rs232_irq,Swap
	XREF SpinIrq0,SpinIrq1,SpinIrq2,SpinIrq3,SpinIrq4,SpinUp,SpinDown
	XREF SpinTicker
	XREF DSPirq,I2CIrq
	XREF Trap1

exception_vecs:
	DC.L	STKTOP	; initial stack pointer
	DC.L	START	; initial execution address
	DC.L	buss_error,address_error				; 0x08
	DC.L	illegal,zero_div,chk_instr,trapv_instr	; 0x10
	DC.L	privilege,trace,line_a,line_f			; 0x20
	DC.L	reserved,reserved,reserved,unitialized	; 0x30
	DCB.L	8,reserved								; 0x40
	DC.L	spurious,autovector,autovector			; 0x60
	DC.L	autovector
	DCB.L	4,autovector				; 0x70
	DC.L	Swap						; 0x80
	DC.L	Trap1						; 0x84
	DCB.L	14,trap_instr				; 0x88
	DCB.L	16,reserved					; 0xc0
	DC.L	_rs232_irq					; 0x100
	DCB.L	127,user_vecs				; 0x104
	DCB.L	8,user_vecs					; 0x300
	DCB.L	8,user_vecs					; 0x320
	DCB.L	2,user_vecs					; 0x340
	DCB.L	46,user_vecs				; 0x348
